module ActiveRedis
  module Attributes

    module InstanceMethods

      ##
      # Save the current record to redis store.
      #
      # @return [true,false]
      #
      def save
        valid? ? (set_attributes(attributes) && true) : false
      end

      ##
      # Reformat object.
      #
      # @return [String]
      #
      def inspect
        a = keys.reject{|k,v| k == :id }.map do |attribute,type|
          %Q{#{attribute}=#{send(attribute).inspect}}
        end
        %Q{#<#{self.class.name} id=#{self.id} #{a.join(' ')}>}
      end

      ##
      # Remove the instance from redis store.
      # Remove all relations to another objects.
      #
      # @return [void]
      #
      def remove
        blto = self.class.instance_variable_get(:@belongs_to)
        (blto || []).each do |rel|
          handle_relation relation_key: "#{rel.to_s.singularize}_id",
            action: :del
        end
        connection.del *attribute_paths
      end

      alias :delete :remove
      alias :destroy :remove

      ##
      # Get the current attributes.
      #
      # @return [Hash]
      #
      def attributes
        get_attributes
      end

      ##
      # @see ActiveRedis#keys
      #
      def keys
        self.class.keys
      end

      ##
      # @see ActiveRedis#connection
      #
      def connection
        self.class.connection
      end

      ##
      # Implement ActiveModel validation support.
      #
      def read_attribute_for_validation key 
        attributes[key]
      end


      module_function

      ##
      # Create getter/setter/bool methods by attributes.
      # This method should be called in a constructor.
      #
      # @return [void]
      #
      def create_attribute_methods
        self.class.instance_eval do
          keys.each do |attribute,type|
            define_method(attribute) do
              attributes[attribute]
            end 

            define_method("#{attribute}=") do |arg|
              attributes[attribute] = format_attribute(attribute,arg.to_s)
            end 

            define_method("#{attribute}?") do
              !attributes[attribute].to_s.empty?
            end 
          end 
        end 
      end

      ##
      # Get attributes from redis store.
      #
      # @param reload [true,false] Force reload from redis.
      #
      # @return [Hash]
      #
      def get_attributes reload: false
        if reload or not @h
          h = (@h and @h[:id]) ? connection.hgetall(basename) : {}
          attribute_hash = {}
          unless h.empty?
            keys.each do |k,v|
              attribute_hash[k.to_sym] = format_attribute(k.to_sym, h[k.to_s])
            end
          end
          @h = attribute_hash
        end
        @h
      end

      ##
      # Casting objects into original values.
      # The type whould be defined at class method "keys"
      #
      # @see ActiveRedis.keys
      #
      # @param key [String,Symbol]
      # @param value [Object]
      #
      # @return [Object] Casted value.
      #
      def format_attribute key, value
        case keys[key.to_sym].to_s
        when /Integer/
          value.to_i
        when /Array/
          connection.lrange("#{basename}:#{key}",0,-1)
        when /Hash/
          connection.hgetall("#{basename}:#{key}")
        when /Time/
          Time.parse(value) rescue nil
        else value
        end
      end

      ##
      # Save given hash to redis server.
      # Refresh updated_at timestamp.
      # Unless id isn't set, get the next primary key.
      #
      # @param arg [Hash]
      #
      # @return [void]
      #
      def set_attributes arg
        remove
        arg[:updated_at] = Time.now.to_s
        arg[:id] ||= connection.incr("#{self.class.name}_id")
        arg.each do |k,v|
          case v
          when Array
            v.each do |v|
              connection.lpush "#{basename}:#{k}", v
            end
          when Hash
            connection.mapped_hmset "#{basename}:#{k}", v
          else
            handle_relation relation_key: k, value: v, action: :add
            connection.hset basename, k, v
          end
        end
        get_attributes(reload: true)
      end

      ##
      # Redis basename.
      #
      # @return [String]
      #
      def basename
        "#{self.class.name.downcase.pluralize}:#{self.id}"
      end

      ##
      # Get attribute paths by keys.
      # Complex types needs to be stored at another path.
      #
      # @return [Array]
      #
      def attribute_paths
        keys.map do |k,v|
          case v.to_s
          when /Hash/, /Array/ then "#{basename}:#{k}"
          else basename
          end
        end.uniq
      end

    end

    module ClassMethods

      ##
      # Fromatted class name.
      #
      # @return [String]
      #
      def cname
        self.name.downcase.pluralize
      end

      ##
      # Create a new instance and save it to redis store.
      #
      # @return [ActiveRedis::Base]
      #
      def create attributes
        a = self.new(attributes)
        a.save
        a
      end

      ##
      # Get all objects.
      #
      # @return [Array<ActiveRedis>]
      #
      def all
        connection.keys("#{cname}:*").map do |key|
          self.new({id: key[/#{cname}:(\d+)/,1].to_i},reload: true)
        end
      end

      ##
      # Redis base connection.
      #
      # @return [Redis]
      #
      def connection
        ActiveRedis.redis
      end

      ##
      # @param id [Integer]
      #
      # @return [ActiveRedis::Base]
      #
      def find id
        a = connection.hgetall("#{cname}:#{id}")
        self.new({id: id},reload: true) unless a.empty?
      end

      ##
      # @see where
      #
      # @return [ActiveRedis::Base]
      #
      def find_by basename: nil
        if basename
          where(basename: basename).first
        end
      end

      ##
      # @params basename [String]
      #
      # @return [Array]
      #
      def where basename: nil
        if basename
          a = connection.smembers(basename)
          a.map{ |entry| self.find(entry) }
        else
          []
        end
      end

      ##
      # Set the keys/attributs for an ActiveRedis::Base child.
      #
      # Example:
      #
      #   keys title: String,
      #     number: Integer
      #
      def keys hash=nil
        hash ? (@keys ||= hash) : @keys
      end


      extend self

    end
  end
end
